#pragma once

#include <limits>

#include "_project.h"
#include "winlogger.h"

//prepared statement name prefix
#define UNIQUE_PREP_PREFIX "PREP_STMT_"

/**
* Connection manager.
* Establishes a connection to postgresql DB
*/

#undef max
#undef min
class PG_LIBPQTYPES_API ConnectionManager
{
	typedef unsigned long long TLongLong;

	const std::string connInfo_;
	PGconn* conn_;
	WinLogger* logger_;
	
	//used to generate unique prepared statement name
	static TLongLong val_;
	bool autoReconnect_;

public:
	/**
	* Constructor.
	* @param connInfo connection info
	* @param logger logger
	*/
	ConnectionManager(const std::string &connInfo, WinLogger* logger, bool autoReconnect = true)
																	: 
																	connInfo_(connInfo), 
																	logger_(logger), 
																	conn_(NULL)	,
																	autoReconnect_(autoReconnect)
	{;}
	~ConnectionManager() 
	{
		closeConnection();
	}
	/**
	* Establishes a connection.
	* If a connection was established early it returns old connection if the connection lost it try to established new connection
	*/
	PGconn* getConnection();
	/**
	* Switch Connection Manager to wotk in auto reconnect way.
	* If the option set to true a connection is re-establishing autmatically otherwise a broken connection exception is broken.
	*/
	void setAutoReconnect(bool value) { autoReconnect_ = value; };
	void closeConnection();
	void getUniquePrepStmtName( /*out*/ std::string &newStmtName);
	/**
	*  Checks if connection is established.
	* @return true in the success
	*/
	inline const bool isConnected();
	void Connect();
protected:
	void checkConnection();
};

inline const bool ConnectionManager::isConnected() 
{ 
	return PQstatus(conn_) == CONNECTION_OK; 
};