#pragma once

#include <iostream>
#include <algorithm>
#include <numeric>
#include <boost/ptr_container/ptr_vector.hpp>
#include <iterator>

template <class T> class multi_array
{
	typedef boost::ptr_vector< boost::nullable<T> > DataContainer;
public:
	typedef T ElementType;
	typedef size_t size_type;

	typedef typename DataContainer::iterator iterator;
	typedef typename DataContainer::const_iterator const_iterator;

	class multi_array_ref
	{
		friend multi_array;

		size_type temp_index;
		size_type index_cursor;

		multi_array &mainArray_;

		multi_array_ref(multi_array &mainArray, size_type idx)
			: mainArray_(mainArray)
			, index_cursor(1)
			, temp_index(idx)
		{
		}
	public:
		multi_array_ref &operator[](size_type idx)
		{
			/*if (index_cursor <= 0 || idx >= mainArray_.dimension[index_cursor])
			{
			throw std::out_of_range("");//todo add message
			}*/
			temp_index = temp_index * mainArray_.dimension[index_cursor++] + idx;

			return *this;
		}

		T& operator= (const T &val)
		{
			T& x = this->operator T&();
			x = val;
			return x;
		}
		operator T&()
		{
			//if (index_cursor != dimensions) throw "bad dimensions number";
			T& retval = mainArray_.array_data[temp_index];
			return retval;
		}
		T *operator->()
		{
			return &mainArray_.array_data[temp_index];
		}
		bool isNull() const
		{
			//if (index_cursor != dimensions) throw "bad dimensions number";
			return mainArray_.array_data.is_null(temp_index);
		}
		void replace(T *newVal) const
		{
			DataContainer::iterator offset = mainArray_.array_data.begin() + temp_index;
			//if (index_cursor != dimensions) throw "bad dimensions number";
			mainArray_.array_data.replace(offset, newVal);
		}
	protected:
		multi_array_ref &operator=(const multi_array_ref &);
		multi_array_ref(const multi_array_ref &);
	};

protected:

	size_type dimensions;
	boost::ptr_vector< boost::nullable<T> > array_data;
	std::vector<size_type> dimension;//todo optimize(this can be replaced with scoped_array)

public:
	multi_array():
	  dimensions(0)
	  {}
	  multi_array (size_type dim, ...)
	  : dimensions(0)
	  {
		  va_list args;
		  va_start(args, dim);
		  set_dimensions_l(dim, args);
		  va_end(args);
	  }
	  multi_array_ref operator[](size_type x)
	  {
		  return multi_array_ref(*this, x);
	  }
	  template<class Iterator>
	  void resize_it(const Iterator b, const Iterator e)
	  {
		  std::copy(b, e, std::back_inserter(dimension));

		  initArrayData();

		  dimensions = dimension.size();
	  }
	  void resize(int x, ...)
	  {
		  va_list args;
		  va_start(args, x);
		  set_dimensions_l(x, args);
		  va_end(args);
	  }

	  static bool is_null(iterator it) { return boost::is_null(it); };

	  size_t get_num_dims()
	  {
		  return dimension.end() - dimension.begin();
	  }
	  size_type get_dim(const size_type idx)
	  {
		return dimension[idx];
	  }
	  template<class Iterator>
	  void get_dims(const Iterator outIter)
	  {
		  std::copy(dimension.begin(), dimension.end(), outIter);

		  size_type elems = 0;
		  elems = std::accumulate(dimension.begin(), dimension.end(), 1, std::multiplies<size_type>());
	  }
	  void replace(const iterator &it, T* elem)
	  {
		  array_data.replace(it, elem);
	  }
	  iterator begin() { return array_data.begin(); }
	  iterator end() { return array_data.end(); }
	  const_iterator begin() const { return array_data.begin(); }
	  const_iterator end() const { return array_data.end(); }

protected:
	void set_dimensions_l(int dim, va_list args)
	{
		dimension.resize(dim);
		dimensions = dimension.size();

		for(int i = 0; i < dimensions; i++) 
			dimension[i] = va_arg( args, int);

		initArrayData();
	}
	void initArrayData()
	{
		size_type elems = 0;
		elems = std::accumulate(dimension.begin(), dimension.end(), 1, std::multiplies<size_type>());

		array_data.resize(elems);
	}


};

