#pragma once

#include "libpq-fe.h"

#include "PG_LibPQTypes.h"
#include <map>
#include <hash_map>
#include <string>
#include "pqtypesimport.h"
 
// for logger stuff
const char StrComma[] = ",";
const char StrRightBrace[] = "}"; 
const char StrLeftBrace[] = "{";
const char StrSingleQuote[] = "\'";
const char StrMinus[] = "-";
const char StrColon[] = ":";
const char StrPoint[] = ".";
//
// pointer for PQnoticeReceiver
typedef void (*UsrNoticeReceiver) (const char *msg);
typedef void (*PQnoticeReceiver) (void *arg, const ::PGresult *res);

namespace TypeReg
{
	enum eTypes
	{
		eInt8,
		eInt4,
		eInt2,
		eFloat8,
		eFloat4,
		eString,
		eConstCharStr,
		eUnknownType
	};
}
namespace PGTypeReg
{
	enum ePGTypes
	{
		eUnknownType,
		eBool,
		eInt8,
		eInt4,
		eInt2,
		eFloat8,
		eFloat4,
		eVarchar,
		eFloat8arr,
		eFloat4arr,
		eInt8arr,
		eInt4arr,
		eDate,
		eTmestamp,
		eMACAddr,
		eINET
	};

	typedef std::map<std::string, ePGTypes> TPGTypeReg;
	typedef std::map<std::string, TypeReg::eTypes> TTypeReg;
	extern TPGTypeReg g_PGTypeReg;
	extern TTypeReg g_TypeReg;

	ePGTypes extern PG_LIBPQTYPES_API getPGTypeByStr(const std::string& strType);
	TypeReg::eTypes extern PG_LIBPQTYPES_API getTypeByStr(const std::string& strType);
	std::string extern PG_LIBPQTYPES_API getInternalPGTypeName(const std::string& strType);
};

namespace StringUtils
{
	//String useful functions
	void extern PG_LIBPQTYPES_API trim_right(std::string &source, const std::string &t = " ");
	void extern PG_LIBPQTYPES_API trim_left(std::string &source, const std::string &t = " ");
	void extern PG_LIBPQTYPES_API tokenizeStr(const std::string& str, std::vector<std::string>& tokens, const std::string& delim = ",");
	void extern PG_LIBPQTYPES_API toLower(std::string& str);
	void extern PG_LIBPQTYPES_API toUpper(std::string& str);
	const bool extern PG_LIBPQTYPES_API ci_equal(const char ch1, const char ch2);
	const size_t extern PG_LIBPQTYPES_API ci_find(const std::string& str1, const std::string& str2);
	const void extern PG_LIBPQTYPES_API transformStringDate(std::string& str);
	struct PostgresTypeToStr
	{
		template<class T>
		std::string operator()(T* t)
		{
			std::string s = "Not supported convertion for :";
			s += typeid(t).name();
			return s;
		};
		template<class T>
		std::string operator()(T t)
		{
			std::stringstream ss;
			ss << t;
			return ss.str();
		};
		template<>
		std::string operator()(PGtimestamp ts)
		{ return operator()(&ts); };
		template<>
		std::string operator()(PGtimestamp *ts)
		{
			std::stringstream ss;
			ss << StrSingleQuote;
			ss << ts->date.year << StrMinus << ts->date.mon + 1 << StrMinus << ts->date.mday;
			ss << " " << ts->time.hour << StrColon << ts->time.min << StrColon << ts->time.sec << StrPoint << ts->time.usec / 1000;
			ss << StrSingleQuote;
			return ss.str();
		}
		template<>
		std::string operator()(PGmacaddr ts)
		{ return operator()(&ts); };
		template<>
		std::string operator()(PGmacaddr* ts)
		{ std::stringstream os;
			os << StrSingleQuote;
			os << ts->a << StrColon << ts->b << StrColon << ts->c;
			os << StrColon << ts->d << StrColon << ts->e << StrColon << ts->f;
			os << StrSingleQuote;
			return os.str();
		}
		template<>
		std::string operator()(PGinet ts)
		{ return operator()(&ts); };
		template<>
		std::string operator()(PGinet* ts)
		{ return "'PGinet type(TODO fix output function for this type)'"; }
		template<>
		std::string operator()(PGdate ts)
		{ return operator()(&ts); };
		template<>
		std::string operator()(PGdate* ts)
		{ std::stringstream os;
			os << StrSingleQuote;
			os << ts->year << StrColon << ts->mon + 1 << StrColon << ts->mday;
			os << StrSingleQuote;
			return os.str();
		}
		std::string operator()(char ts)
		{ return operator()(&ts); };
		template<>
		std::string operator()(char* ts)
		{ std::stringstream os;
			os << ts;
			return os.str();
		}
	};
};
namespace DateTimeUtils
{
	std::string extern PG_LIBPQTYPES_API currentDate(void);
};
namespace LibPQExp
{
	PG_LIBPQTYPES_API char* _PQerrorMessage(const PGconn *conn);
	PG_LIBPQTYPES_API void _PQclear(PGresult *res);
	PG_LIBPQTYPES_API char* _PQresultErrorMessage(const PGresult *res);
	PG_LIBPQTYPES_API ExecStatusType _PQresultStatus(const PGresult *res);
	PG_LIBPQTYPES_API int _PQntuples(const PGresult *res);
	PG_LIBPQTYPES_API int _PQnfields(const PGresult *res);
	PG_LIBPQTYPES_API int _PQgetisnull(const PGresult *res, int tup_num, int field_num);
	PG_LIBPQTYPES_API PGresult* _PQprepare(PGconn *conn, const char *stmtName, const char *query, int nParams, const Oid *paramTypes);
	PG_LIBPQTYPES_API int _PQtypesRegister(PGconn* conn);
	PG_LIBPQTYPES_API char* _PQgetErrorField(int fieldcode);
	PG_LIBPQTYPES_API size_t _PQescapeStringConn (PGconn *conn,	char *to, const char *from, size_t length, int *error);

};
struct NoticeDispatcher
{
	typedef std::map<std::string, UsrNoticeReceiver>TPtrContainer;
	typedef TPtrContainer::iterator					iterator;
	typedef TPtrContainer::const_iterator			const_iterator;
	typedef TPtrContainer::size_type				size_type; 
	static TPtrContainer							elements_;
	// this class is not Copyable nor Assignable 
	void operator=(const NoticeDispatcher&);
	NoticeDispatcher() {;}
public:
	// of course detructor is virtual
	virtual ~NoticeDispatcher() {;}
	// one way of adding new elements
	static void registerForNotice(const std::string& str)
	{
		elements_.insert(std::make_pair(str, (UsrNoticeReceiver)NULL));
	};
	static void subcribeForNotice(const std::string& str, UsrNoticeReceiver rc)
	{
		iterator i = elements_.find(str);
		if (i != elements_.end()) 
			(*i).second = rc;
	};
	static void unSubscribeForNotice(const std::string& str)
	{
		if (elements_.empty()) return;
		iterator i = elements_.find(str);
		if (i != elements_.end())
			NoticeDispatcher::remove(i);
	};
	static void remove(iterator where)
	{
		elements_.erase(where);
	};
	static iterator find(const std::string& str)  
	{
		iterator i = elements_.find(str);
		return i;
	};
	//count the elements 
	static size_type size(void) 
	{
		return elements_.size();
	};
	static void defaultNoticeProcessor(const char *message)
	{
		std::cerr << "NOTICE:  " << message << std::endl;
	};
	static void noticeReceiver(void *arg, const PGresult *result)
	{
		//const char* sqlstate = PQresultErrorField(result, PG_DIAG_SQLSTATE);
		const char* message = PQresultErrorField(result, PG_DIAG_MESSAGE_PRIMARY);
		NoticeDispatcher::doCall(message);
	};
	static void doCall(const char* msg)
	{
		bool nullFound = false;
		for(iterator i = elements_.begin(); i != elements_.end(); ++i)
		{
			if (!(*i).second)
			{
				if (!nullFound)
				{
					defaultNoticeProcessor(msg);
					nullFound = true;
				}
				continue;
			}
			((*i).second)(msg);
		}
	}
};
